/*************************************************************/
/**                                                         **/
/**                         PSG2MID.c                       **/
/**                                                         **/
/** This program converts PSG sound log files to standard   **/
/** MIDI files                                              **/
/**                                                         **/
/** Copyright (C) Marcel de Kogel 1996                      **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/**                                                         **/
/*************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

static const char Title[]=
"PSG2MID 2.01: PSG to MIDI converter\n"
"Copyright (C) 1996  M. de Kogel\n";

/* Patches used */
#define MelodicProg             80              /* Square Lead */
#define NoiseProg               81              /* Sawwooth Lead */
#define SCCProg                 87              /* Bass+Lead */

/* Don't process a note if it's frequency number is lower than 10 */
#define MIN_FREQ        10

/* MIDI channels used */
#define PSG_VOICE               10              /* 11-13 */
#define NOISE_VOICE             13              /* 14-16 */
#define SCC_VOICE               0               /* 1-5 */
#define OPLL_VOICE              0               /* 1-9 + drums (10) */

#define MAX_ENVSPEED            (5*65536)

/* sizeof(byte)=1, sizeof(word)=2 */
typedef unsigned char byte;
typedef unsigned short word;

/* Ticks per second */
static unsigned timeconstant=120;
/* VBlank interrupt frequency */
static unsigned ifreq;

static word ChannelFreq[16]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
static byte midinote[16]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
static word midipitchbend[16]={8192,8192,8192,8192,
                               8192,8192,8192,8192,
                               8192,8192,8192,8192,
                               8192,8192,8192,8192};
static byte midivolume[16]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
static byte tone_on[16]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
static byte midipatch[16]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
static int delta=0;
static FILE *infile,*outfile;
static unsigned PSG_Reg[14];
static unsigned SCC_Reg[32];
static unsigned OPLL_Reg[0x40];
static int soundenvspeed_toolarge;
static int soundenv;
static int usesoundenv[3];
static int soundenvcounter[3];
static int soundenvspeed;
static int soundoutbusy=0;
static int psg=1;
static int scc=0;
static int opll=0;
static int wheel=1;

static int opll_patches[16] =
{
 90,                            /* User (Polysynth) */
 40,                            /* Violin */
 27,                            /* Electric Guitar (clean) */
 1,                             /* Bright Acoustic Piano */
 73,                            /* Flute */
 71,                            /* Clarinet */
 68,                            /* Oboe */
 56,                            /* Trumpet */
 20,                            /* Reed Organ */
 58,                            /* Tuba */
 50,                            /* Synth Strings 1 */
 6,                             /* Harpsichord */
 11,                            /* Vibraphone */
 38,                            /* Synth Bass 1 */
 34,                            /* Electric Bass (pick) */
 33,                            /* Electric Bass (finger) */
};

struct MidiFreq_struct
{
 byte note;
 word wheel;
};

static struct MidiFreq_struct FreqTable[8192]=
{
 #include "ftable.h"
};

/* Put delta time in the MIDI file */
static void WriteDeltaTime (void)
{
 if (delta<128)
  fputc (delta,outfile);
 else
  if (delta<(128*128))
  {
   fputc ((delta>>7)|0x80,outfile);
   fputc (delta&0x7F,outfile);
  }
  else
  {
   fputc (((delta>>14)&0x7F)|0x80,outfile);
   fputc (((delta>>7)&0x7F)|0x80,outfile);
   fputc (delta&0x7F,outfile);
  }
 delta=0;
}

/* Put a MIDI event in the MIDI file. Check if status byte doesn't change */
static void MidiMessage (byte m1,byte m2,byte m3)
{
 static byte midistatus=0;
 WriteDeltaTime ();
 if (m1!=midistatus)
 {
  fputc (m1,outfile);
  midistatus=m1;
 }
 if (m2<128)
 {
  fputc (m2,outfile);
  if (m3<128)
   fputc (m3,outfile);
 }
 m1&=0xF0;
 if (m1==0xC0 || m1==0xD0 || m1==0xF0)
  midistatus=0;
}

static void Volume (byte channel,byte vol)
{
 if (midivolume[channel]!=vol)
 {
  MidiMessage (0xB0+channel,7,vol);
  midivolume[channel]=vol;
 }
}

static void NoteOff (byte channel)
{
 if (midinote[channel])
 {
  MidiMessage (0x90+channel,midinote[channel],0);
  midinote[channel]=0;
 }
}

static void NoteOn (byte channel,byte note)
{
 if (midinote[channel]!=note)
 {
  NoteOff (channel);
  MidiMessage (0x90+channel,note,100);
  midinote[channel]=note;
 }
}

static void PitchBend (byte channel,word pos)
{
 if (midipitchbend[channel]!=pos && wheel)
 {
  MidiMessage (0xE0+channel,pos&0x7F,(pos>>7)&0x7F);
  midipitchbend[channel]=pos;
 }
}

static void ProgramChange (byte channel,byte patch)
{
 if (midipatch[channel]!=patch)
 {
  MidiMessage (0xC0+channel,patch,255);
  midipatch[channel]=patch;
 }
}

static void SetFreq (byte channel, word freq)
{
 ChannelFreq[channel]=freq;
 if (tone_on[channel])
 {
  if (freq<MIN_FREQ)
   NoteOff (channel);
  else
  {
   NoteOn (channel,FreqTable[freq].note);
   PitchBend (channel,FreqTable[freq].wheel);
  }
 }
}

static void SetVolume (byte channel,int vol)
{
 /* These volume scales are almost certainly incorrect */
 if (vol==0)
  Volume (channel,0);
 else if (vol>0)                /* 2 dB steps */
  Volume (channel,127-(15-vol)*16/3);
 else                           /* 1.5 dB steps */
  Volume (channel,127-(vol+31)*16/4);
}

static void ToneOn (byte channel,byte status)
{
 status=(status)? 0:1;
 if (tone_on[channel]!=status)
 {
  tone_on[channel]=status;
  if (tone_on[channel])
   SetFreq (channel,ChannelFreq[channel]);
  else
   NoteOff (channel);
 }
}

static void SetEnvSpeed (unsigned val)
{
 unsigned tmp;
 val&=0xFFFF;
 if (!val)
  val=1;
 tmp=3663932985U/(val*ifreq);
 if (tmp>MAX_ENVSPEED)
 {
  soundenvspeed_toolarge=tmp-MAX_ENVSPEED;
  tmp=MAX_ENVSPEED;
 }
 else
  soundenvspeed_toolarge=0;
 soundenvspeed=tmp;
}

static void PSGOut (byte r,word v)
{
 if (r>13)
  return;
 soundoutbusy=1;
 switch (r)
 {
  case 0:
  case 2:
  case 4:
   SetFreq ((r>>1)+PSG_VOICE,v+(PSG_Reg[r+1]<<8));
   break;
  case 1:
  case 3:
  case 5:
   v&=0x0F;
   SetFreq ((r>>1)+PSG_VOICE,(v<<8)+PSG_Reg[r-1]);
   break;
  case 6:
   v&=0x1F;
   SetFreq (PSG_VOICE+3,v*8+16);
   SetFreq (PSG_VOICE+4,v*8+16);
   SetFreq (PSG_VOICE+5,v*8+16);
   break;
  case 7:
   ToneOn (0+PSG_VOICE,v&0x01);
   ToneOn (1+PSG_VOICE,v&0x02);
   ToneOn (2+PSG_VOICE,v&0x04);
   ToneOn (3+PSG_VOICE,v&0x08);
   ToneOn (4+PSG_VOICE,v&0x10);
   ToneOn (5+PSG_VOICE,v&0x20);
   break;
  case 8:
  case 9:
  case 10:
   v&=0x1F;
   if ((v&10)==0)
   {
    usesoundenv[r-8]=0;
    SetVolume (r-8+PSG_VOICE,v&0x0F);
    SetVolume (r-8+3+PSG_VOICE,v&0x0F);
   }
   else
   {
    usesoundenv[r-8]=1;
    soundenvcounter[r-8]=0;
    if (soundenv&4)
    {
     SetVolume (r-8+PSG_VOICE,1);
     SetVolume (r-8+3+PSG_VOICE,1);
    }
    else
    {
     SetVolume (r-8+PSG_VOICE,15);
     SetVolume (r-8+3+PSG_VOICE,15);
    }
   }
   break;
  case 11:              /* bits 7-0 vd snelheid vh volumeverloop */
   SetEnvSpeed ((PSG_Reg[12]<<8)+v);
   break;
  case 12:              /* bits 15-8 */
   SetEnvSpeed ((v<<8)+PSG_Reg[11]);
   break;
  case 13:              /* soort volumeverloop */
   v&=0x0F;
   soundenv=v;
   break;
 }
 PSG_Reg[r]=v;
 soundoutbusy=0;
}

static void SCCOut (byte r,word v)
{
 if (r>0x8F)
  return;
 if (r<0x80)
  return;               /* wave change */
 r&=0xF;
 switch (r)
 {
  case 0x00:
  case 0x02:
  case 0x04:
  case 0x06:
  case 0x08:
   SetFreq ((r>>1)+SCC_VOICE,v+(SCC_Reg[r+1]<<8));
   break;
  case 0x01:
  case 0x03:
  case 0x05:
  case 0x07:
  case 0x09:
   v&=0x0F;
   SetFreq ((r>>1)+SCC_VOICE,(v<<8)+SCC_Reg[r-1]);
   break;
  case 0x0A:
  case 0x0B:
  case 0x0C:
  case 0x0D:
  case 0x0E:
   v&=0x0F;
   SetVolume (r-0x0A+SCC_VOICE,v);
   break;
  case 0x0F:
   v^=0xFF;
   ToneOn (SCC_VOICE+0,v&0x01);
   ToneOn (SCC_VOICE+1,v&0x02);
   ToneOn (SCC_VOICE+2,v&0x04);
   ToneOn (SCC_VOICE+3,v&0x08);
   ToneOn (SCC_VOICE+4,v&0x10);
   break;
 }
 SCC_Reg[r]=v;
}

static byte GetDrumVolume (byte ch)
{
 byte v=0;
 switch (ch)
 {
  case 1: v=OPLL_Reg[0x37]>>4; break;
  case 2: v=OPLL_Reg[0x38]&0x0F; break;
  case 3: v=OPLL_Reg[0x38]>>4; break;
  case 4: v=OPLL_Reg[0x37]&0x0F; break;
  case 5: v=OPLL_Reg[0x36]&0x0F; break;
 }
 v=127-(15-v)*3;
 return v;
}

static void OPLLDrums (byte value)
{
 static byte drum_notes[5]= { 42,49,47,40,36 };
 static byte oldvalue=0x80;
 if (oldvalue==0x80 || ((oldvalue&1)!=(value&1)))
  MidiMessage (0x99,drum_notes[0],(value&1)? GetDrumVolume(1):0);
 if (oldvalue==0x80 || ((oldvalue&2)!=(value&2)))
  MidiMessage (0x99,drum_notes[1],(value&2)? GetDrumVolume(2):0);
 if (oldvalue==0x80 || ((oldvalue&4)!=(value&4)))
  MidiMessage (0x99,drum_notes[2],(value&4)? GetDrumVolume(3):0);
 if (oldvalue==0x80 || ((oldvalue&8)!=(value&8)))
  MidiMessage (0x99,drum_notes[3],(value&8)? GetDrumVolume(4):0);
 if (oldvalue==0x80 || ((oldvalue&16)!=(value&16)))
  MidiMessage (0x99,drum_notes[4],(value&16)? GetDrumVolume(5):0);
 oldvalue=value;
}

static void OPLL_SetFreq (byte ch,word freq)
{
 SetFreq (ch+OPLL_VOICE,freq+4096);
}

static void OPLL_ToneOn (byte ch,byte status)
{
 ToneOn (ch+OPLL_VOICE,(status)? 0:1);
}

void OPLLOut(byte reg, byte value)
{
 int ch;
 static byte OPLLMode=0;
 if (reg>63)
  return;
 OPLL_Reg[reg]=value;
 ch=reg&0x0F;
 switch (reg>>4)
 {
  case 0:
   if(ch==0x0E)
   {
    OPLLMode=value&32;
    if (OPLLMode)
    {
     NoteOff (OPLL_VOICE+6);
     NoteOff (OPLL_VOICE+7);
     NoteOff (OPLL_VOICE+8);
     OPLLDrums (value&0x1F);
    }
    else
     OPLLDrums (0);
   }
   /* register 0x0F: */
   /* bit 0: 1=Sound is off */
   /* bit 3: 1=Sample mode */
   break;
  case 1:
   if (ch>8) break;
   if (OPLLMode==0 || ch<6)
    OPLL_SetFreq (ch,((OPLL_Reg[reg+0x10]&0x0F)<<8)|value);
   break;
  case 2:
   if (ch>8) break;
   if (OPLLMode==0 || ch<6)
   {
    OPLL_ToneOn (ch,value&0x10);
    OPLL_SetFreq (ch,((value&0x0F)<<8)|OPLL_Reg[reg-0x10]);
   }
   break;
  case 3:
   if (ch>8) break;
   if (OPLLMode==0 || ch<6)
   {
    ProgramChange (ch+OPLL_VOICE,opll_patches[value>>4]);
    SetVolume (ch+OPLL_VOICE,15-(value&0xF));
   }
   break;
 }
}

static void Interrupt_Routine (void)
{
 int i,tmp,counter;
 int volume=0;
 if (soundoutbusy)
  return;
 soundoutbusy=1;
 for (i=0;i<3;++i)
 {
  tmp=soundenvcounter[i]>>16;
  soundenvcounter[i]=(soundenvcounter[i]+soundenvspeed)&0x7FFFFFFF;
  counter=soundenvcounter[i]>>16;
  if (tmp!=counter && usesoundenv[i])
  {
   switch (soundenv)
   {
    case 0: case 1: case 2: case 3: case 9:
     if (counter>=30)
     {
      usesoundenv[i]=0;
      counter=30;
     }
     volume=30-counter;
     break;
    case 4: case 5: case 6: case 7: case 15:
     if (counter>30)
     {
      usesoundenv[i]=0;
      counter=0;
     }
     volume=counter;
     break;
    case 8:
     while (counter>30)
     {
      counter-=30;
      soundenvcounter[i]-=(30<<16);
     }
     volume=30-counter;
     if (soundenvspeed_toolarge)
      volume=24;
     break;
    case 10:
     while (counter>60)
     {
      counter-=60;
      soundenvcounter[i]-=(60<<16);
     }
     volume=abs(30-counter);
     if (soundenvspeed_toolarge)
      volume=24;
     break;
    case 11:
     if (counter>30)
     {
      usesoundenv[i]=0;
      counter=0;
     }
     volume=30-counter;
     break;
    case 12:
     while (counter>30)
     {
      counter-=30;
      soundenvcounter[i]-=(30<<16);
     }
     volume=counter;
     if (soundenvspeed_toolarge)
      volume=24;
     break;
    case 13:
     if (counter>30)
     {
      usesoundenv[i]=0;
      counter=30;
     }
     volume=counter;
     break;
    case 14:
     while (counter>60)
     {
      counter-=60;
      soundenvcounter[i]-=(60<<16);
     }
     volume=30-abs(30-counter);
     if (soundenvspeed_toolarge)
      volume=24;
     break;
   }
   if (usesoundenv[i] || volume)
   {
    SetVolume (i+PSG_VOICE,-volume-1);
    SetVolume (i+3+PSG_VOICE,-volume-1);
   }
   else
   {
    SetVolume (i+PSG_VOICE,0);
    SetVolume (i+3+PSG_VOICE,0);
    NoteOff (i+PSG_VOICE);
    NoteOff (i+3+PSG_VOICE);
   }
  }
 }
 soundoutbusy=0;
}

static void WriteMidiHeader (void)
{
 int tempo;
 int i;
 /* "MThd" */
 fputc ('M',outfile);
 fputc ('T',outfile);
 fputc ('h',outfile);
 fputc ('d',outfile);
 /* Header data length (6) */
 fputc (0,outfile);
 fputc (0,outfile);
 fputc (0,outfile);
 fputc (6,outfile);
 /* File format (0) */
 fputc (0,outfile);
 fputc (0,outfile);
 /* Number of tracks (1) */
 fputc (0,outfile);
 fputc (1,outfile);
 /* Time constant */
 fputc ((timeconstant>>8)&0xFF,outfile);
 fputc (timeconstant&0xFF,outfile);

 /* "MTrk" */
 fputc ('M',outfile);
 fputc ('T',outfile);
 fputc ('r',outfile);
 fputc ('k',outfile);
 /* Track length (filled in PatchMidiHeader() */
 fputc (0,outfile);
 fputc (0,outfile);
 fputc (0,outfile);
 fputc (0,outfile);
 /* Delta time 0 */
 fputc (0,outfile);
 /* Tempo */
 tempo=500000*timeconstant*2/(ifreq*4);
 fputc (0xFF,outfile);
 fputc (0x51,outfile);
 fputc (3,outfile);
 fputc ((tempo>>16)&0xFF,outfile);
 fputc ((tempo>>8)&0xFF,outfile);
 fputc (tempo&0xFF,outfile);
 /* Put a MIDI controller 0 change in the file to change to user bank 1 */
 if (psg)
 {
  for (i=PSG_VOICE;i<PSG_VOICE+6;++i)
   MidiMessage (0xB0+i,0,1);
 }
 /* Load the correct patches */
 if (psg)
 {
  for (i=PSG_VOICE;i<PSG_VOICE+3;++i)
   ProgramChange (i,MelodicProg);
  for (i=NOISE_VOICE;i<NOISE_VOICE+3;++i)
   ProgramChange (i,NoiseProg);
 }
 if (scc)
  for (i=SCC_VOICE;i<SCC_VOICE+5;++i)
   ProgramChange (i,SCCProg);
 if (opll)
  for (i=OPLL_VOICE;i<OPLL_VOICE+9;++i)
   ProgramChange (i,opll_patches[0]);
 /* Set volume to 0 */
 for (i=PSG_VOICE;i<PSG_VOICE+6;++i)
  MidiMessage (0xB0+i,7,0);
 if (scc)
  for (i=SCC_VOICE;i<SCC_VOICE+5;++i)
   MidiMessage (0xB0+i,7,0);
 if (opll)
 {
  for (i=OPLL_VOICE;i<OPLL_VOICE+9;++i)
   MidiMessage (0xB0+i,7,0);
  /* Set drum channel volume to max. */
  MidiMessage (0xB9,7,127);
 } 
}

/* Get file length and put track length in file at offset 18 */
static void PatchMidiHeader (void)
{
 long tracklen;
 fseek (outfile,0,SEEK_END);
 tracklen=ftell(outfile)-22;
 fseek (outfile,18,SEEK_SET);
 fputc (tracklen>>24,outfile);
 fputc ((tracklen>>16)&0xFF,outfile);
 fputc ((tracklen>>8)&0xFF,outfile);
 fputc (tracklen&0xFF,outfile);
}

int main (int argc,char *argv[])
{
 static char *options[] =
 {
  "psg", "nopsg", "scc", "noscc", "opll", "noopll", "ifreq",
  "nowheel", "wheel",
  NULL
 };
 char PSGSoundHeader[]="PSG";
 int a,b,i,j,misparm;
 byte buffer[128];
 char *szInfile=NULL;
 char *szOutfile=NULL;
 printf (Title);
 a=0;
 misparm=0;
 ifreq=0;
 for (i=1;i<argc;++i)
 {
  if (argv[i][0]!='-')
  {
   switch (a)
   {
    case 0: szInfile=argv[i]; break;
    case 1: szOutfile=argv[i]; break;
    default: misparm=1;
   }
   ++a;
  }
  else
  {
   for (j=0;options[j];++j)
    if (!strcmp(argv[i]+1,options[j])) break;
   switch (j)
   {
    case 0: psg=1; break;
    case 1: psg=0; break;
    case 2: scc=1; break;
    case 3: scc=0; break;
    case 4: opll=1; break;
    case 5: opll=0; break;
    case 6: ++i;
            if (i<argc) ifreq=atoi(argv[i]);
            else misparm=1;
            break;
    case 7: wheel=0; break;
    case 8: wheel=1; break;
    default: misparm=1;
   }
  }
 }
 if (misparm || !szInfile || !szOutfile || (scc && opll))
 {
  printf ("Usage: %s [options] <PSG file> <midi file>\n",argv[0]);
  printf ("Available options are:\n"
          "-psg            - Enable PSG (default)\n"
          "-nopsg          - Disable PSG\n"
          "-scc            - Enable SCC\n"
          "-noscc          - Disable SCC (default)\n"
          "-opll           - Enable OPLL\n"
          "-noopll         - Disable OPLL (default)\n"
          "-ifreq <value>  - Set interrupt frequency (default=50Hz)\n"
          "-nowheel        - Disable pitch wheel events\n"
          "-wheel          - Enable pitch wheel events (default)\n"
          "Enabling both SCC and OPLL is not possible\n");
  return 1;
 }
 infile=fopen (szInfile,"rb");
 if (infile==NULL)
 {
  printf ("Cannot open %s\n",szInfile);
  return 2;
 }
 fread (buffer,1,16,infile);
 buffer[3]='\0';
 if (strcmp (buffer,PSGSoundHeader))
 {
  printf ("%s is not a PSG sound file\n",argv[1]);
  return 4;
 }
 if (buffer[4]>0x10)
 {
  printf ("Sound file version not supported\n");
  return 5;
 }
 if (!ifreq)
 {
  if (buffer[4]>=0x10)
   ifreq=buffer[5];
  if (ifreq<10 || ifreq>1000)
   ifreq=50;
 }
 outfile=fopen (szOutfile,"wb");
 if (outfile==NULL)
 {
  printf ("Cannot open %s\n",szOutfile);
  return 3;
 }
 WriteMidiHeader ();
 while ((a=fgetc(infile))!=EOF)
 {
  switch (a)
  {
   case 0xFF:
    for (i=0;i<4;++i)
    {
     ++delta;
     Interrupt_Routine ();
    }
    break;
   case 0xFE:
    a=fgetc (infile);
    for (i=0;i<a*4;++i)
    {
     ++delta;
     Interrupt_Routine ();
    }
    break;
   default:
    b=fgetc (infile);
    if (a<0x10)
     if (psg) PSGOut (a,b);
    if (a>=0x10 && a<0xA0)
     if (scc) SCCOut (a-0x10,b);
    if (a>=0xA0 && a<0xE0)
     if (opll) OPLLOut (a-0xA0,b);
    break;
  }
 }
 /* All notes off */
 if (psg) PSGOut (7,0xFF);
 if (scc) SCCOut (0xF,0);
 if (opll)
 {
  OPLLOut (0x0E,0);
  for (i=0;i<9;++i)
   OPLLOut (0x20+i,0);
 }
 /* End of track */
 MidiMessage (0xFF,0x2F,0x00);
 /* Put track length in file */
 PatchMidiHeader ();
 fclose (infile);
 fclose (outfile);
 return 0;
}

